use strict;
use Encode;
undef $/;
my $dir = 'c:/jobs/propeller/';
my $files = 0;
my @opcodes = qw/wrbyte rdbyte wrword rdword wrlong rdlong hubop clkset cogid coginit cogstop locknew lockret lockset lockclr ror rol shr shl
								 rcr rcl sar rev mins maxs min max movs movd movi jmpret jmp call ret test and andn or xor muxc muxnc muxz muxnz add sub cmp
								 addabs subabs sumc sumnc sumz sumnz mov neg abs absneg negc negnc negz negnz cmps cmpsx addx subx cmpx adds subs addsx
								 subsx cmpsub djnz tjnz tjz waitpeq waitpne waitcnt waitvid nop byte word long/;
my %freq = map {$_ => 0} (@opcodes);
my $op_pat = join('|', @opcodes);
my @headers = qw/con obj dat pub pri var/;
my $hd_pat = join('|', @headers);
my $in_dat = 0;
my $nlines = 0;
my $nimm = 0;
my $nwait = 0;

opendir DIR, $dir;
while (my $file = readdir DIR) {
	$file = "$dir$file";
	next unless -f $file && $file =~ m/\.spin$/i;
	print "$file\n";
	$files++;
	open SPN, "<$file";
	my $spin = <SPN>;
	close SPN;
	if (substr($spin, 0, 2) eq chr(0xff).chr(0xfe)) {
		$spin = decode('UTF-16', $spin);
		print "UTF-16\n"
	}
	$spin = uncomment($spin);
	my @lines = split(/[\n\r]+/, $spin);
	foreach my $line (@lines) {
		if ($line =~ m/^($hd_pat)($|\W)/i) {
			#print "$line\n";
			my $header = $1;
			$in_dat = ($header =~ m/DAT/i)
		} elsif ($in_dat) {
			if ($line =~ m/\s+($op_pat)\s/i) {
				my $opcode = lc($1);
				if ($opcode =~ m/waitpeq|waitpne/) {
					$nwait++;
					$nimm++ if $line =~ m/,\s*\#/
				}
				$freq{$opcode}++;
				$nlines++;
				#print "$line\n";
				#print "$opcode\n"
			}
		}
	}
}
closedir DIR;
print "Opcode frequencies for $nlines lines of PASM code.\n_______________________\n\n";
foreach my $opcode (sort {$freq{$b} <=> $freq{$a}} (@opcodes)) {
	print "$opcode:", ' ' x (8 - length($opcode)),  sprintf("%6.1d\n", $freq{$opcode})
}

print "$nimm/$nwait";

sub uncomment {
	my @source = split(/[\n\r]+/, shift());
	my $returnsource = '';
	my $nestcomment = 0;
	my $supercomment = 0;
	foreach my $line (@source) {
		my $linecomment = 0;
		my $instring = 0;
		my $chr0 = 0;
		while ($line =~ m/(\{\{|\}\}|\"|\{|\}|\')/g) {
			my $chr1 = pos($line);
			my $delim = $1;
			$chr1 -= length($delim);
			if ($linecomment | $nestcomment | $supercomment) {
				$returnsource .= ' ' x ($chr1 - $chr0)
			} else {
				$returnsource .= substr($line, $chr0, $chr1 - $chr0)
			}
			if ($delim =~ m/\{|\}|\'/ && $instring == 0 || ($linecomment | $nestcomment | $supercomment)) {
				$returnsource .= ' ' x length($delim)
			} else {
				$returnsource .= $delim
			}
			if ($delim eq "{{" && ($instring | $linecomment) == 0) {
				if ($nestcomment) {
					$nestcomment += 2
				} else {
					$supercomment = 1
				}
			} elsif ($delim eq "}}" && ($instring | $linecomment) == 0) {
				if ($nestcomment) {
					$nestcomment -= 2;
					$nestcomment = 0 if $nestcomment < 0
				} else {
					$supercomment = 0
				}
			} elsif ($delim eq "{" && ($instring | $linecomment | $supercomment) == 0) {
				$nestcomment++
			} elsif ($delim eq "}" && $nestcomment && ($instring | $linecomment | $supercomment) == 0) {
				$nestcomment--;
			} elsif ($delim eq "'" && ($instring | $nestcomment | $supercomment) == 0) {
				$linecomment = 1
			} elsif ($delim eq '"' && ($linecomment | $nestcomment | $supercomment) == 0) {
				$instring = 1 - $instring;
			} else {
			}
			$chr0 = $chr1 + length($delim);
		}
		if ($linecomment | $nestcomment | $supercomment) {
			$returnsource .= ' ' x (length($line) - $chr0)
		} else {
			$returnsource .= substr($line, $chr0) unless $linecomment | $nestcomment | $supercomment
		}
		$returnsource .= "\n" unless $nestcomment | $supercomment
	}
	return $returnsource
}


